/*
	APL Restful ARDBC Plugin
    Copyright 2015-2018  LJ LongWing - A Programming Legacy (http://remedylegacy.com)

	Licensed under the Apache License, Version 2.0 (the "License");
	you may not use this file except in compliance with the License.
	You may obtain a copy of the License at

		http://www.apache.org/licenses/LICENSE-2.0

	Unless required by applicable law or agreed to in writing, software
	distributed under the License is distributed on an "AS IS" BASIS,
	WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
	See the License for the specific language governing permissions and
	limitations under the License.
 */

package com.apl;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.UnsupportedEncodingException;
import java.net.MalformedURLException;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import org.apache.commons.codec.binary.Base64;
import org.apache.commons.lang.StringUtils;

import com.bmc.arsys.api.ARException;
import com.bmc.arsys.api.ARServerUser;
import com.bmc.arsys.api.ArithmeticOrRelationalOperand;
import com.bmc.arsys.api.AttachmentValue;
import com.bmc.arsys.api.Constants;
import com.bmc.arsys.api.DataType;
import com.bmc.arsys.api.Entry;
import com.bmc.arsys.api.Form;
import com.bmc.arsys.api.OperandType;
import com.bmc.arsys.api.QualifierInfo;
import com.bmc.arsys.api.RelationalOperationInfo;
import com.bmc.arsys.api.Timestamp;
import com.bmc.arsys.api.Value;
import com.bmc.arsys.pluginsvr.plugins.ARPluginContext;
import com.bmc.arsys.pluginsvr.plugins.ARVendorField;

public class Functions {
	public static Map<String, Object> parseQual(QualifierInfo qual, List<ARVendorField> fields, ARPluginContext context) {
		Map<String, Object> fieldList = new TreeMap<>();

		if (qual == null || qual.getOperation() == Constants.AR_COND_OP_NONE) {
			return fieldList;
		}

		QualifierInfo leftQual = null;
		QualifierInfo rightQual = null;

		if (qual.getOperation() != Constants.AR_COND_OP_REL_OP) {
			leftQual = qual.getLeftOperand();
			rightQual = qual.getRightOperand();

			fieldList.putAll(parseQual(leftQual, fields, context));
			fieldList.putAll(parseQual(rightQual, fields, context));
		}

		switch (qual.getOperation()) {
		case Constants.AR_COND_OP_NONE:
			break;
		case Constants.AR_COND_OP_FROM_FIELD:
			break;
		case Constants.AR_COND_OP_AND:
			break;
		case Constants.AR_COND_OP_OR:
			break;
		case Constants.AR_COND_OP_NOT:
			break;
		case Constants.AR_COND_OP_REL_OP:
			RelationalOperationInfo operations = qual.getRelationalOperationInfo();

			ArithmeticOrRelationalOperand leftOperand = operations.getLeftOperand();
			Object leftOperandValue = leftOperand.getValue();

			ArithmeticOrRelationalOperand rightOperand = operations.getRightOperand();
			Object rightOperandValue = rightOperand.getValue();

			OperandType leftOperandType = leftOperand.getType();
			OperandType rightOperandType = rightOperand.getType();

			String left = processOperand(leftOperandType, leftOperandValue, fields, context);
			String right = processOperand(rightOperandType, rightOperandValue, fields, context);

			String fieldName = "";
			String fieldValue = "";

			// if both left and right are values, we have a possible boolean situation here
			if (leftOperandType == OperandType.VALUE && rightOperandType == OperandType.VALUE) {
				if (!left.equals(right))
					fieldList.put("booleanResult", "false");
			}
			
			if (leftOperandType == OperandType.FIELDID)
				fieldName = left;
			else if (leftOperandType == OperandType.VALUE)
				fieldValue = left;

			if (rightOperandType == OperandType.FIELDID)
				fieldName = right;
			else if (rightOperandType == OperandType.VALUE)
				fieldValue = right;

			if (!fieldName.equals("") && !fieldValue.equals("")) {
				fieldList.put(fieldName, fieldValue);
			}

			/*
			 * int relOp = operations.getOperation(); switch (relOp) { case
			 * Constants.AR_REL_OP_EQUAL: break; case
			 * Constants.AR_REL_OP_GREATER: output(
			 * "Unhandled RelationalOperation: > is not currently supported on field '"
			 * + fieldName + "'", context); break; case
			 * Constants.AR_REL_OP_GREATER_EQUAL: output(
			 * "Unhandled RelationalOperation: >= is not currently supported on field '"
			 * + fieldName + "'", context); break; case
			 * Constants.AR_REL_OP_LESS: output(
			 * "Unhandled RelationalOperation: < is not currently supported on field '"
			 * + fieldName + "'", context); break; case
			 * Constants.AR_REL_OP_LESS_EQUAL: output(
			 * "Unhandled RelationalOperation: <= is not currently supported on field '"
			 * + fieldName + "'", context); break; case
			 * Constants.AR_REL_OP_NOT_EQUAL: output(
			 * "Unhandled RelationalOperation: != is not currently supported on field '"
			 * + fieldName + "'", context); break; case
			 * Constants.AR_REL_OP_LIKE: output(
			 * "Unhandled RelationalOperation: LIKE is not currently supported on field '"
			 * + fieldName + "'", context); break; default: output(
			 * "Unhandled relOp: " + relOp, context); }
			 */
			break;
		default:
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Unhandled parseQual Conditional Operation: ");
		}

		return fieldList;
	}

	static boolean processLocalFilter(QualifierInfo qual, List<ARVendorField> fields, List<String> fieldLocalFilterList,
			Entry entry, ARPluginContext context) {
		boolean valid = false;

		if (qual == null || qual.getOperation() == Constants.AR_COND_OP_NONE) {
			valid = true;
			return valid;
		}

		QualifierInfo leftQual = null;
		QualifierInfo rightQual = null;

		boolean left = false;
		boolean right = false;

		if (qual.getOperation() != Constants.AR_COND_OP_REL_OP) {
			leftQual = qual.getLeftOperand();
			rightQual = qual.getRightOperand();

			left = processLocalFilter(leftQual, fields, fieldLocalFilterList, entry, context);
			right = processLocalFilter(rightQual, fields, fieldLocalFilterList, entry, context);
		}

		// 'Request ID' = "12345"

		switch (qual.getOperation()) {
		case Constants.AR_COND_OP_NONE:
			break;
		case Constants.AR_COND_OP_FROM_FIELD:
			break;
		case Constants.AR_COND_OP_AND:
			if (left && right)
				valid = true;
			break;
		case Constants.AR_COND_OP_OR:
			if (left || right)
				valid = true;
			break;
		case Constants.AR_COND_OP_NOT:
			if (!left)
				valid = true;
			break;
		case Constants.AR_COND_OP_REL_OP:
			RelationalOperationInfo operations = qual.getRelationalOperationInfo();

			ArithmeticOrRelationalOperand leftOperand = operations.getLeftOperand();
			Object leftOperandValue = leftOperand.getValue();
			OperandType leftOperandType = leftOperand.getType();

			ArithmeticOrRelationalOperand rightOperand = operations.getRightOperand();
			Object rightOperandValue = rightOperand.getValue();
			OperandType rightOperandType = rightOperand.getType();

			String leftOp = processOperand(leftOperandType, leftOperandValue, fields, context);
			String rightOp = processOperand(rightOperandType, rightOperandValue, fields, context);

			String fieldName = "";
			int fieldID = 0;
			Value fieldValue = null;

			if (leftOperandType == OperandType.FIELDID) {
				fieldName = leftOp;
				fieldID = Integer.parseInt(leftOperandValue.toString());
			} else if (leftOperandType == OperandType.VALUE)
				fieldValue = (Value) leftOperandValue;

			if (rightOperandType == OperandType.FIELDID) {
				fieldName = rightOp;
				fieldID = Integer.parseInt(rightOperandValue.toString());
			} else if (rightOperandType == OperandType.VALUE)
				fieldValue = (Value) rightOperandValue;

			Value entryFieldValue = entry.get(fieldID);

			if (entryFieldValue == null)
				entryFieldValue = new Value();

			if (!fieldName.equals("") 
					&& fieldID > 0
					//&& fieldValue.getDataType().toInt() == entryFieldValue.getDataType().toInt() //in the case of null values, the two datatypes don't match
					&& fieldLocalFilterList.contains(fieldName)) {
				int relOp = operations.getOperation();
				switch (relOp) {
				case Constants.AR_REL_OP_EQUAL:
					switch (fieldValue.getDataType().toInt()) {
					case Constants.AR_DATA_TYPE_NULL:
						if (entryFieldValue.getDataType().toInt() == Constants.AR_DATA_TYPE_NULL) {
							valid = true;
						}
						break;
					case Constants.AR_DATA_TYPE_CHAR:
						if (fieldValue.toString().equals(entryFieldValue.toString())) {
							valid = true;
						}
						break;
					case Constants.AR_DATA_TYPE_INTEGER:
						if (fieldValue.getIntValue() == entryFieldValue.getIntValue()) {
							valid = true;
						}
						break;
					case Constants.AR_DATA_TYPE_TIME:
						if (fieldValue.getIntValue() == entryFieldValue.getIntValue()) {
							valid = true;
						}
						break;
					default:
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Unhandled Data Type: " + fieldValue.getDataType().toInt() + " - " + fieldValue.toString());
					}
					break;
				case Constants.AR_REL_OP_GREATER:
					switch (fieldValue.getDataType().toInt()) {
					case Constants.AR_DATA_TYPE_NULL:
						break;
					case Constants.AR_DATA_TYPE_CHAR:
						if (entryFieldValue.toString().compareTo(fieldValue.toString()) > 0)
							valid = true;
						break;
					case Constants.AR_DATA_TYPE_INTEGER:
						if (entryFieldValue.getIntValue() > fieldValue.getIntValue())
							valid = true;
						break;
					case Constants.AR_DATA_TYPE_TIME:
						if (entryFieldValue.getIntValue() > fieldValue.getIntValue())
							valid = true;
						break;
					default:
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Unhandled Data Type: " + fieldValue.getDataType().toInt() + " - " + fieldValue.toString());
					}
					break;
				case Constants.AR_REL_OP_GREATER_EQUAL:
					switch (fieldValue.getDataType().toInt()) {
					case Constants.AR_DATA_TYPE_NULL:
						break;
					case Constants.AR_DATA_TYPE_CHAR:
						if (entryFieldValue.toString().compareTo(fieldValue.toString()) >= 0)
							valid = true;
						break;
					case Constants.AR_DATA_TYPE_INTEGER:
						if (entryFieldValue.getIntValue() >= fieldValue.getIntValue())
							valid = true;
						break;
					case Constants.AR_DATA_TYPE_TIME:
						if (entryFieldValue.getIntValue() >= fieldValue.getIntValue())
							valid = true;
						break;
					default:
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Unhandled Data Type: " + fieldValue.getDataType().toInt() + " - " + fieldValue.toString());
					}
					break;
				case Constants.AR_REL_OP_LESS:
					switch (fieldValue.getDataType().toInt()) {
					case Constants.AR_DATA_TYPE_NULL:
						break;
					case Constants.AR_DATA_TYPE_CHAR:
						if (entryFieldValue.toString().compareTo(fieldValue.toString()) < 0)
							valid = true;
						break;
					case Constants.AR_DATA_TYPE_INTEGER:
						if (entryFieldValue.getIntValue() < fieldValue.getIntValue())
							valid = true;
						break;
					case Constants.AR_DATA_TYPE_TIME:
						if (entryFieldValue.getIntValue() < fieldValue.getIntValue())
							valid = true;
						break;
					default:
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Unhandled Data Type: " + fieldValue.getDataType().toInt() + " - " + fieldValue.toString());
					}
					break;
				case Constants.AR_REL_OP_LESS_EQUAL:
					switch (fieldValue.getDataType().toInt()) {
					case Constants.AR_DATA_TYPE_NULL:
						break;
					case Constants.AR_DATA_TYPE_CHAR:
						if (entryFieldValue.toString().compareTo(fieldValue.toString()) <= 0)
							valid = true;
						break;
					case Constants.AR_DATA_TYPE_INTEGER:
						if (entryFieldValue.getIntValue() <= fieldValue.getIntValue())
							valid = true;
						break;
					case Constants.AR_DATA_TYPE_TIME:
						if (entryFieldValue.getIntValue() <= fieldValue.getIntValue())
							valid = true;
						break;
					default:
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Unhandled Data Type: " + fieldValue.getDataType().toInt() + " - " + fieldValue.toString());
					}
					break;
				case Constants.AR_REL_OP_NOT_EQUAL:
					switch (fieldValue.getDataType().toInt()) {
					case Constants.AR_DATA_TYPE_NULL:
						if (entryFieldValue.getDataType().toInt() != Constants.AR_DATA_TYPE_NULL) {
							valid = true;
						}
						break;
					case Constants.AR_DATA_TYPE_CHAR:
						if (!fieldValue.toString().equals(entryFieldValue.toString()))
							valid = true;
						break;
					case Constants.AR_DATA_TYPE_INTEGER:
						if (fieldValue.getIntValue() != entryFieldValue.getIntValue())
							valid = true;
						break;
					case Constants.AR_DATA_TYPE_TIME:
						if (fieldValue.getIntValue() != entryFieldValue.getIntValue())
							valid = true;
						break;
					default:
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Unhandled Data Type: " + fieldValue.getDataType().toInt() + " - " + fieldValue.toString());
					}
					break;
				case Constants.AR_REL_OP_LIKE:
					switch (fieldValue.getDataType().toInt()) {
					case Constants.AR_DATA_TYPE_NULL:
						break;
					case Constants.AR_DATA_TYPE_CHAR:
						String curFieldValue = fieldValue.toString();
						curFieldValue = curFieldValue.replace(".", "\\.");
						curFieldValue = curFieldValue.replace("%", ".*");
						curFieldValue = curFieldValue.replace("_", ".");
						if (entryFieldValue.toString().matches(curFieldValue))
							valid = true;
						break;
					case Constants.AR_DATA_TYPE_INTEGER:
						break;
					case Constants.AR_DATA_TYPE_TIME:
						break;
					default:
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Unhandled Data Type: " + fieldValue.getDataType().toInt() + " - " + fieldValue.toString());
					}

					break;
				default:
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Unhandled relOp: " + relOp);
				}
				break;
			} else {
				valid = true;
			}
			break;
		default:
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Unhandled processLocalFilter Conditional Operation: " + qual.getOperation());
		}

		return valid;
	}

	private static String processOperand(OperandType operandType, Object operandValue, List<ARVendorField> fields,
			ARPluginContext context) {
		String s = "";

		if (operandType.equals(OperandType.FIELDID)) {
			for (ARVendorField field : fields) {
				if (field.getFieldId() == Integer.parseInt(operandValue.toString())) {
					s = field.getName();
					break;
				}
			}
		} else if (operandType.equals(OperandType.VALUE)) {
			Value v = (Value) operandValue;

			int dataType = v.getDataType().toInt();
			switch (dataType) {
			case Constants.AR_DATA_TYPE_NULL:
				s = "";
				break;
			case Constants.AR_DATA_TYPE_CHAR:
				s = v.toString();
				break;
			case Constants.AR_DATA_TYPE_INTEGER:
				s = v.toString();
				break;
			case Constants.AR_DATA_TYPE_TIME:
				s = v.toString();
				break;
			default:
				s = "Unhandled Data Type: " + dataType + " - " + v.toString();
			}
			// } else if (operandType.equals(OperandType.ARITHMETIC_OP)) {

		} else {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Unhandled Operand Type: " + operandType);
		}

		return s;
	}

	static String addElementPrefix(String current, String change) {
		String newValue = current + change + ":";

		if (newValue.startsWith(":"))
			newValue = newValue.substring(1);

		return newValue;
	}

	static String removeElementSuffix(String current, String change) {
		int curLen = current.length();
		int chgLen = change.length() + 1;

		String newValue = current.substring(0, curLen - chgLen);

		return newValue;
	}

	static Map<String, Object> processEntry(Entry entry, List<ARVendorField> fields, Map<String, APLField> aplFieldMap) {
		Map<String, Object> fieldList = new TreeMap<String, Object>();

		if (entry != null) {
			for (java.util.Map.Entry<Integer, Value> entryMap : entry.entrySet()) {
				String fieldName = "";
				Object fieldValueS = null;
				int fieldID = entryMap.getKey();
				Value fieldValue = entryMap.getValue();

				for (ARVendorField vendorField : fields) {
					int vendorFieldID = vendorField.getFieldId();
					if (vendorFieldID == fieldID) {
						fieldName = vendorField.getName();
						break;
					}
				}

				if (fieldValue != null && fieldValue.getDataType().toInt() != Constants.AR_DATA_TYPE_NULL) {
					//System.out.println("Field: " + fieldName);
					if (fieldValue.getDataType() == DataType.ATTACHMENT) {
						//System.out.println("Field is Attachment");
						Object valueObject = fieldValue.getValue();
						AttachmentValue attField = (AttachmentValue)valueObject;

						byte[] attBytes = attField.getContent();
						if (attBytes != null) {
							//System.out.println("Bytes are not null");
							boolean bypassEncoding = false;
							if (aplFieldMap != null && aplFieldMap.containsKey(fieldName)) {
								//System.out.println("Found the field in the field map");
								bypassEncoding = aplFieldMap.get(fieldName).bypassEncoding();
								//System.out.println("Setting bypassEncoding: " + bypassEncoding);
							} else {
								//System.out.println((aplFieldMap != null) + ":" + (aplFieldMap.containsKey(fieldName)) + ":" + aplFieldMap.size());
							}
							//System.out.println("After If/Else");
							
							if (bypassEncoding) {
								//try {
									//fieldValueS = new String(attBytes, "UTF8");
								//} catch (UnsupportedEncodingException e) {}
									fieldValueS = attBytes;
							} else {
								fieldValueS = java.util.Base64.getEncoder().encodeToString(attBytes);
							}
						}
						String fileName = attField.getName();
						if (fileName.contains("/"))
							fileName = fileName.substring(fileName.lastIndexOf("/")+1);
						if (fileName.contains("\\"))
							fileName = fileName.substring(fileName.lastIndexOf("\\")+1);
						fieldList.put(fieldName + ":name", fileName);
						fieldList.put(fieldName + ":size", attField.getOriginalSize());
					} else if (fieldValue.getDataType() == DataType.TIME) {
						Object valueObject = fieldValue.getValue();
						Timestamp dateTime = (Timestamp)valueObject;
						DateFormat df = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss");
						if (dateTime != null)
							fieldValueS = df.format(dateTime.toDate());
					} else {
						fieldValueS = fieldValue.toString();
					}
				}

				if (fieldName != null && fieldValueS != null && !fieldName.isEmpty()) {
					fieldList.put(fieldName, fieldValueS);
				}
			}
		}

		return fieldList;
	}

	public static String replaceTemplateValues(String template, Map<String, Object> fieldNameValueMap, List<String> optionalList
			, Map<String, APLField> aplFieldMap, String templateFieldDelimiter) {
		RestfulARDBCPlugin.attByteList.clear();
		// loop through the new entry field values for replacement in the
		// template

		if (template != null) {
			//loop through the template and remove any optional null fields
			for (String optionalField : optionalList) {
				if (!fieldNameValueMap.containsKey(optionalField)) {
					template = template.replaceAll("(?m)^.*\\Q" + templateFieldDelimiter + "\\E" + optionalField + "\\Q" + templateFieldDelimiter + "\\E.*$", "");
					template = template.replaceAll("(?m)^[ \t]*\r?\n", "");
				}
			}

			//look through the field Name/Value map and replace any template values with the field values
			for (String fieldName : fieldNameValueMap.keySet()) {
				Object fieldValue = null;

				if (fieldNameValueMap.containsKey(fieldName)) {
					fieldValue = fieldNameValueMap.get(fieldName);
				}

				if (fieldValue == null)
					fieldValue = "";

				boolean bypassEncoding = false;
				if (aplFieldMap != null && aplFieldMap.containsKey(fieldName))
					bypassEncoding = aplFieldMap.get(fieldName).bypassEncoding();
				
				if (!bypassEncoding) {
					if (template.startsWith("{")) {
						if (fieldValue instanceof String && ((String)fieldValue).contains("\\")) {
							fieldValue = ((String)fieldValue).replaceAll("\\\\", "\\\\\\\\");
						}
						if (fieldValue instanceof String && ((String)fieldValue).contains("\"")) {
							fieldValue = ((String)fieldValue).replaceAll("\"", "\\\\\"");
						}
						if (fieldValue instanceof String && ((String)fieldValue).contains("/")) {
							fieldValue = ((String)fieldValue).replaceAll("/", "\\\\/");
						}
						if (fieldValue instanceof String && ((String)fieldValue).contains("\n")) {
							fieldValue = ((String)fieldValue).replaceAll("\n", "\\\\n");
						}
						if (fieldValue instanceof String && ((String)fieldValue).contains("\r")) {
							fieldValue = ((String)fieldValue).replaceAll("\r", "\\\\r");
						}
						if (fieldValue instanceof String && ((String)fieldValue).contains("\t")) {
							fieldValue = ((String)fieldValue).replaceAll("\\t", "\\\\t");
						}
					} else if (template.startsWith("<")) {
						if (fieldValue instanceof String && ((String)fieldValue).contains("&"))
							fieldValue = ((String)fieldValue).replaceAll("&", "&amp;");
						if (fieldValue instanceof String && ((String)fieldValue).contains("<"))
							fieldValue = ((String)fieldValue).replaceAll("<", "&lt;");
						if (fieldValue instanceof String && ((String)fieldValue).contains(">"))
							fieldValue = ((String)fieldValue).replaceAll(">", "&gt;");
						if (fieldValue instanceof String && ((String)fieldValue).contains("\""))
							fieldValue = ((String)fieldValue).replaceAll("\"", "&quot;");
						if (fieldValue instanceof String && ((String)fieldValue).contains("'"))
							fieldValue = ((String)fieldValue).replaceAll("'", "&apos;");
					}
				}

				if (fieldValue instanceof String)
					template = template.replace(templateFieldDelimiter + fieldName + templateFieldDelimiter, (String)fieldValue);
				else if (fieldValue instanceof byte[]){
					RestfulARDBCPlugin.attByteList.add((byte[]) fieldValue);
					int attByteSize = RestfulARDBCPlugin.attByteList.size();
					template = template.replace(templateFieldDelimiter + fieldName + templateFieldDelimiter, "byteListLoc" + (attByteSize-1) + "byteListLoc");
				}
			}

			//replace any remaining field template values with ""
			template = template.replaceAll("(?m)\\Q" + templateFieldDelimiter + "\\E.*\\Q" + templateFieldDelimiter + "\\E", "");
		} else {
			template = "";
		}
		return template;
	}

	static int getErrorFieldID(List<ARVendorField> fields) {
		int fieldID = 0;

		for (ARVendorField field : fields) {
			if (field.getName().equals("ErrorOutput")) {
				fieldID = field.getFieldId();
				continue;
			}
		}

		return fieldID;
	}

	static String getErrorString(List<ARVendorField> fields, List<Entry> restResultsList) {
		int errorField = Functions.getErrorFieldID(fields);
		Entry entry = restResultsList.get(0);
		String message = "";
		for (java.util.Map.Entry<Integer, Value> entryMap : entry.entrySet()) {
			if (entryMap.getKey() == errorField) {
				message = entryMap.getValue().toString();
			}
		}

		return message;
	}

	static String encodeURL(String urlS) throws URISyntaxException, MalformedURLException {
		String encodedUrl = null;
		URL url = new URL(urlS);
		String urlProtocol = url.getProtocol();
		String urlUserInfo = url.getUserInfo();
		String urlHost = url.getHost();
		int urlPort = url.getPort();
		String urlPath = url.getPath();
		String urlQuery = url.getQuery();
		String urlRef = url.getRef();

		if (urlRef != null && !urlRef.equals("")) {
			urlQuery += "#" + urlRef;
			urlRef = null;
		}

		URI uri = new URI(urlProtocol, urlUserInfo, urlHost, urlPort, urlPath, urlQuery, urlRef);
		encodedUrl = uri.toASCIIString();

		return encodedUrl;
	}

	static Entry deserializeEntryObject (ARPluginContext context, String serializedEntry) {
		Entry entry = null;

		try {
			byte b[] = null;
			b = java.util.Base64.getDecoder().decode(serializedEntry);
			ByteArrayInputStream bi = new ByteArrayInputStream(b);
			ObjectInputStream si = new ObjectInputStream(bi);
			entry = (Entry) si.readObject();
		} catch (Exception e) {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Error converting String to Entry Object CacheEntry: " + e.getMessage());
			return null;
		}

		return entry;
	}

	static String serializeEntryObject(Entry entry) throws IOException {
		ByteArrayOutputStream bo = new ByteArrayOutputStream();
		ObjectOutputStream so = new ObjectOutputStream(bo);
		so.writeObject(entry);
		so.flush();
		return java.util.Base64.getEncoder().encodeToString(bo.toByteArray());
	}

	static String setUserName (ARPluginContext context) {
		String userName = context.getUser();
		if (userName == null || userName.isEmpty())
			userName = "Remedy Application Service";

		return userName;
	}

	static boolean getDebug (List<ARVendorField> fields, Entry entry, boolean alwaysLogDetails) {
		if (alwaysLogDetails)
			return true;
		else {
			for (ARVendorField field : fields) {
				if (field.getName().equals("Debug")) {
					int fieldID = field.getFieldId();

					Value fieldValue = entry.get(fieldID);

					if (fieldValue != null && fieldValue.toString() != null && !fieldValue.toString().isEmpty())
						return true;
				}
			}
		}

		return false;
	}

	static boolean getDebug (Map<String, Object> fieldNameValueMap, boolean alwaysLogDetails) {
		if (alwaysLogDetails)
			return true;
		else if (fieldNameValueMap != null) {
			String s = (String)fieldNameValueMap.get("Debug");
			if ((s != null && !s.equals("")))
				return true;
		}

		return false;
	}

	static String generateEntryID(int number, long transID) {
		String entryID = String.valueOf(transID);
		int eidLength = entryID.length();
		int padNeeded = 15 - eidLength;
		String newNumber = String.valueOf(number);
		if (padNeeded > 0)
			newNumber = StringUtils.leftPad(String.valueOf(number), padNeeded, "0");
		entryID += newNumber; 
		
		return entryID;
	}

	static int lengthOfInt(int i) {
		int length = 0;
		long temp = 1;
		while (temp <= i) {
			length++;
			temp *= 10;
		}
		return length;
	}

	public static byte[] processInputStream(ARPluginContext context, InputStream is) {
		byte[] bytes = null;
		ByteArrayOutputStream buffer = new ByteArrayOutputStream();

		int nRead;
		byte[] data = new byte[4096];

		try {
			while ((nRead = is.read(data, 0, data.length)) != -1) {
				buffer.write(data, 0, nRead);
			}
			buffer.flush();
			bytes = buffer.toByteArray();
		} catch (IOException e) {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR,
					"Error closing InputStream during getAttachment: " + e.getMessage());
		}
		

		return bytes;
	}
	
	static InputStream stringToIS (String s, String encoding) {
		InputStream is = null;
		
		if (encoding != null && !encoding.isEmpty()) {
			try {
				is = new ByteArrayInputStream(s.getBytes(encoding));
			} catch (UnsupportedEncodingException e) {
				is = new ByteArrayInputStream(s.getBytes());
			}
		} else {
			is = new ByteArrayInputStream(s.getBytes());
		}
		
		return is;
	}

	static List<Integer> enumSkipElementLevels(String s) {
		List<Integer> skipLevelList = new ArrayList<Integer>();
		String[] tokens = s.split(",");
		for (String token : tokens) {
			try {
				skipLevelList.add(Integer.valueOf(token.trim()));
			} catch (NumberFormatException e) {}
		}
		return skipLevelList;
	}
	
	static String getCharacterSet(String s) {
		String charset = null;

		if (s != null) {
			String[] values = s.split(";"); // values.length should be 2
			if (values != null) {
				for (String value : values) {
					value = value.trim();

					if (value.toLowerCase().startsWith("charset=")) {
						charset = value.substring("charset=".length());
					}
				}
			}
		}

		return charset;
	}
	
	static Map<String, String> generateHeaderElements(Map<String, List<String>> headerMap, ARPluginContext context, boolean debug) {
		if (debug)
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Processing returned headers");
		Map<String, String> returnMap = new HashMap<String, String>();
		if (headerMap != null) {
			for (String headerKey : headerMap.keySet()) {
				List<String> headerValueList = headerMap.get(headerKey);
				StringBuilder sb = new StringBuilder();
				if (headerValueList != null) {
					for (String headerValue : headerValueList) {
						if (sb.toString().isEmpty()) {
							sb.append(headerValue);
						} else {
							sb.append(",");
							sb.append(headerValue);
						}
					}
				}
				String headerValues = sb.toString();
				if (headerKey != null && !headerKey.isEmpty() && headerValues != null && !headerValues.isEmpty()) {
					headerKey = "Header:" + headerKey;
					if (debug)
						Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_INFO, "Adding headerKey|Value: " + headerKey + "|" + headerValues);
					returnMap.put(headerKey, headerValues);
				}
			}
		}
		
		return returnMap;
	}
	
	static List<Map<String, String>> processArrays(Map<String, List<String>> arrayElementMap, Map<String, String> globalElementMap, Map<String, String> recordElementMap, List<String> fieldNameList, String arrayDelimiter) {
		for(String elementName : arrayElementMap.keySet()) {
			if (elementName == null || elementName.isEmpty())
				continue;

			String value = "";
			
			List<String> valueList = arrayElementMap.get(elementName);
			for (String curValue : valueList) {
				if (value.isEmpty())
					value = curValue;
				else
					value += arrayDelimiter + curValue;
			}
			
			if (globalElementMap == null)
				globalElementMap = new HashMap<String, String>();
			if (recordElementMap == null)
				recordElementMap = new HashMap<String, String>();
			
			if (globalElementMap.containsKey(elementName))
				globalElementMap.put(elementName, value);
			if (recordElementMap.containsKey(elementName))
				recordElementMap.put(elementName, value);

			//begin processing elements with named values
			if (!fieldNameList.isEmpty()) {
				for (String fieldName : fieldNameList) {
					if (fieldName.startsWith(elementName) && !fieldName.equals(elementName)) {
						String targetElement = fieldName;
						String valueSearchLeft = fieldName.substring(elementName.length() + 1);
						if (valueSearchLeft.contains(":")) {
							String valueSearchRight = valueSearchLeft.substring(valueSearchLeft.indexOf(":")+1);
							valueSearchLeft = valueSearchLeft.substring(0, valueSearchLeft.indexOf(":"));
							int valueIndex = valueList.indexOf(valueSearchLeft);
							String searchElement = elementName.substring(0, elementName.lastIndexOf(":")+1) + valueSearchRight;
							if (valueIndex >= 0) {
								if (arrayElementMap.containsKey(searchElement)) {
									List<String> valueList2 = arrayElementMap.get(searchElement);
									if (valueList2.size() > valueIndex) {
										String targetValue = valueList2.get(valueIndex);
										if (globalElementMap.containsKey(elementName))
											globalElementMap.put(targetElement, targetValue);
										if (recordElementMap.containsKey(elementName))
											recordElementMap.put(targetElement, targetValue);
									}
								}
							}
						} else {
							try {
								int index = Integer.valueOf(valueSearchLeft);
								String targetValue = valueList.get(index);
								if (globalElementMap.containsKey(elementName))
									globalElementMap.put(targetElement, targetValue);
								if (recordElementMap.containsKey(elementName))
									recordElementMap.put(targetElement, targetValue);
							} catch (NumberFormatException e) {
								//System.err.println("Unable to convert " + valueSearchLeft + " to a number");
							}
						}
					}
				}
			}		
		}
		
		List<Map<String, String>> mapList = new ArrayList<Map<String, String>>();
		if (globalElementMap.size() > 0)
			mapList.add(globalElementMap);
		else
			mapList.add(null);
		if (recordElementMap != null && recordElementMap.size() > 0)
			mapList.add(recordElementMap);
		else
			mapList.add(null);

		return mapList;
	}

	public static String base64EncodeString(String s) {
		byte[] byteArray = Base64.encodeBase64(s.getBytes());
		return new String(byteArray);
	}
	
	public static String getFormName(ARPluginContext context, ARServerUser conn, int[] fieldArray, String defaultName) {
		String formName = null;
		
		List<Form> formList = null;
		try {
			formList = conn.getListFormObjects(0, Constants.AR_LIST_SCHEMA_REGULAR | Constants.AR_HIDDEN_INCREMENT, null, fieldArray, null);
		} catch (ARException e) {
			Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Unable to retrieve form name: " + formList);
		}
		
		if (formList != null && formList.size() > 0) {
			if (formList.size() == 1)
				formName = formList.get(0).getName();
			else {
				for (Form form : formList) {
					formName = form.getName();
					if (formName.equals(defaultName)) {
						break;
					}
				}
				if (!formName.equals(defaultName)) {
					Functions.logMessage(context, ARPluginContext.PLUGIN_LOG_LEVEL_ERROR, "Multiple possible forms found for '" + defaultName + "', unable to determine proper form, please manually specify");
					return defaultName;
				}
			}
		}
		return formName;
	}
	
	public static QualifierInfo getQualifierInfo(int fieldID, String value, int relationship) {
		ArithmeticOrRelationalOperand leftOperand = new ArithmeticOrRelationalOperand(fieldID);
		ArithmeticOrRelationalOperand rightOperand = new ArithmeticOrRelationalOperand(new Value(value));
		RelationalOperationInfo relOpInfo = new RelationalOperationInfo(relationship, leftOperand, rightOperand);
		QualifierInfo qual = new QualifierInfo(relOpInfo);
		return qual;
	}
	public static QualifierInfo getQualifierInfo(int fieldID, int value, int relationship) {
		ArithmeticOrRelationalOperand leftOperand = new ArithmeticOrRelationalOperand(fieldID);
		ArithmeticOrRelationalOperand rightOperand = new ArithmeticOrRelationalOperand(new Value(value));
		RelationalOperationInfo relOpInfo = new RelationalOperationInfo(relationship, leftOperand, rightOperand);
		QualifierInfo qual = new QualifierInfo(relOpInfo);
		return qual;
	}
	
	public static QualifierInfo getActiveTableQual(String tableName) {
		QualifierInfo tableQual = Functions.getQualifierInfo(APLConstants.fidTableName, tableName, RelationalOperationInfo.AR_REL_OP_EQUAL);
		QualifierInfo qual = new QualifierInfo(QualifierInfo.AR_COND_OP_AND, APLConstants.statusActiveQual, tableQual);
		return qual;
	}
	
	public static QualifierInfo getActiveGUIDQual(String guid) {
		QualifierInfo guidQual = Functions.getQualifierInfo(APLConstants.fidGuid, guid, RelationalOperationInfo.AR_REL_OP_EQUAL);
		QualifierInfo qual = new QualifierInfo(QualifierInfo.AR_COND_OP_AND, APLConstants.statusActiveQual, guidQual);
		return qual;
	}

	public static void logMessage(ARPluginContext context, int logLevel, String message ) {
		if (context != null) {
			context.logMessage(context.getPluginInfo(), logLevel, message);
		} else {
			System.out.println("Context is null, unable to log in proper log file:");
			System.out.println(logLevel + ": " + message);
		}
	}
}
